import asyncio
import io
import os
import sys
import traceback
from types import (
    TracebackType,
)
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    BinaryIO,
    Tuple,
    cast,
)

from async_generator import (
    asynccontextmanager,
)
import cloudpickle


def get_subprocess_command(
    child_r: int, child_w: int, parent_pid: int
) -> Tuple[str, ...]:
    from . import _child

    return (
        sys.executable,
        "-m",
        _child.__name__,
        "--parent-pid",
        str(parent_pid),
        "--fd-read",
        str(child_r),
        "--fd-write",
        str(child_w),
    )


def pickle_value(value: Any) -> bytes:
    serialized_value = cloudpickle.dumps(value)
    # mypy doesn't recognize that this line produces a bytes type.
    return len(serialized_value).to_bytes(4, 'big') + serialized_value  # type: ignore


def read_exactly(stream: BinaryIO, num_bytes: int) -> bytes:
    buffer = io.BytesIO()
    bytes_remaining = num_bytes
    while bytes_remaining > 0:
        data = stream.read(bytes_remaining)
        if data == b"":
            raise ConnectionError("Got end of stream")
        buffer.write(data)
        bytes_remaining -= len(data)

    return buffer.getvalue()


def receive_pickled_value(stream: BinaryIO) -> Any:
    len_bytes = read_exactly(stream, 4)
    serialized_len = int.from_bytes(len_bytes, "big")
    serialized_result = read_exactly(stream, serialized_len)
    return cloudpickle.loads(serialized_result)


class RemoteTraceback(Exception):

    def __init__(self, tb: str) -> None:
        self.tb = tb

    def __str__(self) -> str:
        return self.tb


class RemoteException(Exception):
    def __init__(self, exc: BaseException, tb: TracebackType) -> None:
        self.tb = (
            f'\n""" (exception from process: {os.getpid()})\n'
            f"{''.join(traceback.format_exception(type(exc), exc, tb))}"
            '"""'
        )
        self.exc = exc

    def __reduce__(self) -> Any:
        """
        Trick the `pickle` module into recreating this as the original
        exception when the value gets unpickled.
        """
        return rebuild_exc, (self.exc, self.tb)


def rebuild_exc(exc, tb):  # type: ignore
    exc.__cause__ = RemoteTraceback(tb)
    return exc


def cleanup_tasks(*tasks: 'asyncio.Future[Any]') -> AsyncContextManager[None]:
    """
    Context manager that ensures that all tasks are properly cancelled and awaited.

    The order in which tasks are cleaned is such that the first task will be
    the last to be cancelled/awaited.

    This function **must** be called with at least one task.
    """
    return cast(
        AsyncContextManager[None],
        _cleanup_tasks(*tasks),
    )


# mypy recognizes this decorator as being untyped.
@asynccontextmanager  # type: ignore
async def _cleanup_tasks(task: 'asyncio.Future[Any]',
                         *tasks: 'asyncio.Future[Any]',
                         ) -> AsyncIterator[None]:
    try:
        if tasks:
            async with cleanup_tasks(*tasks):
                yield
        else:
            yield
    finally:
        if not task.done():
            task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass



import os
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    TypeVar,
)

if TYPE_CHECKING:
    #
    # This block ensures that typing_extensions isn't needed during runtime and
    # is only necessary for linting.
    #
    import subprocess
    from typing import Mapping, Optional, Sequence, Union

    from .typing_compat import Literal, TypedDict

    try:
        # mypy knows this isn't present in <=3.6
        from subprocess import STARTUPINFO  # type: ignore
    except ImportError:
        # STARTUPINFO added in 3.7+
        STARTUPINFO = Any

    PIPE = Literal[subprocess.PIPE]
    DEVNULL = Literal[subprocess.DEVNULL]
    STDOUT = Literal[subprocess.STDOUT]

    SubprocessKwargs = TypedDict(
        'SubprocessKwargs',
        # no: bufsize, universal_newlines, shell, text, encoding and errors
        {
            'stdin': Optional[Union[IO, PIPE, DEVNULL]],
            'stdout': Optional[Union[IO, PIPE, DEVNULL]],
            'stderr': Optional[Union[IO, PIPE, DEVNULL, STDOUT]],

            'limit': Optional[int],

            'preexec_fn': Callable[..., Any],
            'close_fds': Optional[bool],
            'pass_fds': Optional[Sequence[int]],
            'cdw': Optional[os.PathLike],
            'restore_signals': Optional[bool],
            'start_new_session': Optional[bool],
            'env': Optional[Mapping[str, str]],
            'startupinfo': Optional[STARTUPINFO],
            'creationflags': Optional[Union[
                Literal[
                    subprocess.CREATE_NEW_CONSOLE,  # type: ignore
                    subprocess.CREATE_NEW_PROCESS_GROUP,  # type: ignore
                    subprocess.ABOVE_NORMAL_PRIORITY_CLASS,  # type: ignore
                    subprocess.BELOW_NORMAL_PRIORITY_CLASS,  # type: ignore
                    subprocess.HIGH_PRIORITY_CLASS,  # type: ignore
                    subprocess.IDLE_PRIORITY_CLASS,  # type: ignore
                    subprocess.NORMAL_PRIORITY_CLASS,  # type: ignore
                    subprocess.REALTIME_PRIORITY_CLASS,  # type: ignore
                    subprocess.CREATE_NO_WINDOW,  # type: ignore
                    subprocess.DETACHED_PROCESS,  # type: ignore
                    subprocess.CREATE_DEFAULT_ERROR_MODE,  # type: ignore
                    subprocess.CREATE_BREAKAWAY_FROM_JOB,  # type: ignore
                ],
                Sequence[Literal[
                    subprocess.CREATE_NEW_CONSOLE,
                    subprocess.CREATE_NEW_PROCESS_GROUP,
                    subprocess.ABOVE_NORMAL_PRIORITY_CLASS,
                    subprocess.BELOW_NORMAL_PRIORITY_CLASS,
                    subprocess.HIGH_PRIORITY_CLASS,
                    subprocess.IDLE_PRIORITY_CLASS,
                    subprocess.NORMAL_PRIORITY_CLASS,
                    subprocess.REALTIME_PRIORITY_CLASS,
                    subprocess.CREATE_NO_WINDOW,
                    subprocess.DETACHED_PROCESS,
                    subprocess.CREATE_DEFAULT_ERROR_MODE,
                    subprocess.CREATE_BREAKAWAY_FROM_JOB,
                ]],
            ]],
        },
        total=False,
    )
else:
    # Ensure this is importable outside of the `TYPE_CHECKING` context so that
    # 3rd party libraries can use this as a type without having to think too
    # much about what they are doing.
    SubprocessKwargs = Dict[str, Any]


TReturn = TypeVar("TReturn")



from abc import (
    ABC,
    abstractmethod,
)
import signal
from typing import (
    Generic,
    Optional,
)

from .state import (
    State,
)
from .typing import (
    TReturn,
)


class ProcessAPI(ABC, Generic[TReturn]):
    sub_proc_payload: bytes

    #
    # State
    #
    @property
    @abstractmethod
    def state(self) -> State:
        ...

    @abstractmethod
    async def update_state(self, value: State) -> None:
        ...

    @abstractmethod
    async def wait_for_state(self, state: State) -> None:
        ...

    #
    # PID
    #
    @property
    @abstractmethod
    def pid(self) -> int:
        ...

    @pid.setter
    def pid(self, value: int) -> None:
        raise NotImplementedError

    @abstractmethod
    async def wait_pid(self) -> int:
        ...

    #
    # Return Value
    #
    @property
    @abstractmethod
    def return_value(self) -> TReturn:
        ...

    @return_value.setter
    def return_value(self, value: TReturn) -> None:
        raise NotImplementedError

    @abstractmethod
    async def wait_return_value(self) -> TReturn:
        ...

    #
    # Return Code
    #
    @property
    @abstractmethod
    def returncode(self) -> int:
        ...

    @returncode.setter
    def returncode(self, value: int) -> None:
        raise NotImplementedError

    @abstractmethod
    async def wait_returncode(self) -> int:
        ...

    #
    # Error
    #
    @property
    @abstractmethod
    def error(self) -> Optional[BaseException]:
        ...

    @error.setter
    def error(self, value: BaseException) -> None:
        raise NotImplementedError

    @abstractmethod
    async def wait_error(self) -> Optional[BaseException]:
        ...

    #
    # Result
    #
    @property
    @abstractmethod
    def result(self) -> TReturn:
        ...

    @abstractmethod
    async def wait_result(self) -> TReturn:
        ...

    #
    # Lifecycle management APIs
    #
    @abstractmethod
    async def wait(self) -> None:
        ...

    @abstractmethod
    async def kill(self) -> None:
        ...

    @abstractmethod
    def terminate(self) -> None:
        ...

    @abstractmethod
    def send_signal(self, sig: signal.Signals) -> None:
        ...



import asyncio
import os
import signal
from typing import (
    Any,
    Callable,
    Optional,
    Sequence,
)

from ._utils import (
    pickle_value,
)
from .abc import (
    ProcessAPI,
)
from .exceptions import (
    ProcessKilled,
)
from .state import (
    State,
)
from .typing import (
    TReturn,
)


class Process(ProcessAPI[TReturn]):
    _pid: Optional[int] = None
    _returncode: Optional[int] = None
    _return_value: TReturn
    _error: Optional[BaseException] = None
    _state: State = State.INITIALIZING

    sub_proc_payload: bytes

    def __init__(
        self, async_fn: Callable[..., TReturn], args: Sequence[Any]
    ) -> None:
        self._async_fn = async_fn
        self._args = args
        self.sub_proc_payload = pickle_value((self._async_fn, self._args))

        self._has_pid = asyncio.Event()
        self._has_returncode = asyncio.Event()
        self._has_return_value = asyncio.Event()
        self._has_error = asyncio.Event()
        self._state_changed = asyncio.Condition()

    def __str__(self) -> str:
        return f"Process[{self._async_fn.__name__}]"

    #
    # State
    #
    @property
    def state(self) -> State:
        """
        Return the current state of the process.
        """
        return self._state

    async def update_state(self, state: State) -> None:
        """
        Update the state of the process.

        .. warning::

            This is an internal API and should not be used in user code.
        """
        async with self._state_changed:
            self._state = state
            self._state_changed.notify_all()

    async def wait_for_state(self, state: State) -> None:
        """
        Block until the process has reached or surpassed the given state.
        """
        if self.state.is_on_or_after(state):
            return

        # We use a loop since there should be a finite number of possible state
        # transitions and thus we should arrived at the desired state within a
        # number of iterations equal to that of the number of possible states.
        for _ in range(len(State)):
            async with self._state_changed:
                await self._state_changed.wait()
                if self.state.is_on_or_after(state):
                    break
        else:
            raise BaseException(
                f"This code path should not be reachable since there are a "
                f"finite number of state transitions.  Current state is "
                f"{self.state}"
            )

    #
    # PID
    #
    @property
    def pid(self) -> int:
        """
        Return the process id of the process

        Raises an `AttributeError` if the process id is not yet available.
        """
        if self._pid is None:
            raise AttributeError("No PID set for process")
        return self._pid

    @pid.setter
    def pid(self, value: int) -> None:
        self._pid = value
        self._has_pid.set()

    async def wait_pid(self) -> int:
        """
        Block until the process id is available.
        """
        await self._has_pid.wait()
        return self.pid

    #
    # Return Value
    #
    @property
    def return_value(self) -> TReturn:
        """
        Return the return value of the proc

        Raises an `AttributeError` if the process has not exited.
        """
        if not hasattr(self, "_return_value"):
            raise AttributeError("No return_value set")
        return self._return_value

    @return_value.setter
    def return_value(self, value: TReturn) -> None:
        self._return_value = value
        self._has_return_value.set()

    async def wait_return_value(self) -> TReturn:
        """
        Block until the return code of the process has been set.

        This will block indefinitely if the process exits with an error.
        """
        await self._has_return_value.wait()
        return self.return_value

    #
    # Return Code
    #
    @property
    def returncode(self) -> int:
        """
        Return the integer return code of the process.

        Raises an `AttributeError` if the process has not exited.
        """
        if self._returncode is None:
            raise AttributeError("No returncode set")
        return self._returncode

    @returncode.setter
    def returncode(self, value: int) -> None:
        self._returncode = value
        self._has_returncode.set()

    async def wait_returncode(self) -> int:
        """
        Block until the return code of the process has been set.
        """
        await self._has_returncode.wait()
        return self.returncode

    #
    # Error
    #
    @property
    def error(self) -> Optional[BaseException]:
        """
        Return the error raised by the process.

        Raises an `AttributeError` if the process has not raised an exception.
        """
        if self._error is None and not hasattr(self, "_return_value"):
            raise AttributeError("No error set")
        return self._error

    @error.setter
    def error(self, value: BaseException) -> None:
        self._error = value
        self._has_error.set()

    async def wait_error(self) -> BaseException:
        """
        Block until the process has an error.

        This will block indefinitely if the process does not throw an exception.
        """
        await self._has_error.wait()
        # mypy is unable to tell that `self.error` **must** be non-null in this
        # case.
        return self.error  # type: ignore

    #
    # Result
    #
    @property
    def result(self) -> TReturn:
        """
        Return the computed result from the process.

        If the process has not finished then raises an `AttributeError`

        If the process raised an exception, that exception is raised.

        Otherwise returns the result of the process.
        """
        if self._error is None and not hasattr(self, "_return_value"):
            raise AttributeError("Process not done")
        elif self._error is not None:
            raise self._error
        elif hasattr(self, "_return_value"):
            return self._return_value
        else:
            raise BaseException("Code path should be unreachable")

    async def wait_result(self) -> TReturn:
        """
        Block until the process has exited, either returning the return value
        if execution was successful, or raising an exception if it failed
        """
        await self.wait_returncode()

        if self.returncode == 0:
            return await self.wait_return_value()
        else:
            raise await self.wait_error()

    #
    # Lifecycle management APIs
    #
    async def wait(self) -> None:
        """
        Block until the process has exited.
        """
        await self.wait_returncode()

        if self.returncode == 0:
            await self.wait_return_value()
        else:
            await self.wait_error()

    async def kill(self) -> None:
        """
        Issue a `SIGKILL` signal to the process.

        This immediately transitions the process state to `FINISHED` and sets
        the error to `ProcessKilled`
        """
        self.send_signal(signal.SIGKILL)
        await self.update_state(State.FINISHED)
        self.error = ProcessKilled("Process terminated with SIGKILL")

    def terminate(self) -> None:
        """
        Issues a `SIGTERM` to the process.
        """
        self.send_signal(signal.SIGTERM)

    def send_signal(self, signum: signal.Signals) -> None:
        """
        Issues the provided signal to the process.
        """
        os.kill(self.pid, signum.value)



import enum


class State(enum.IntEnum):
    """
    Child process lifecycle
    """

    INITIALIZING = 0
    INITIALIZED = 1
    WAIT_EXEC_DATA = 2
    BOOTING = 3
    STARTED = 4
    EXECUTING = 5
    STOPPING = 6
    FINISHED = 7

    def is_next(self, other: "State") -> bool:
        return other == self + 1

    def is_on_or_after(self, other: "State") -> bool:
        return self >= other

    def is_before(self, other: "State") -> bool:
        return self < other



import asyncio
import logging
import os
import signal
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    AsyncIterator,
    Callable,
    Optional,
    cast,
)

from async_generator import (
    asynccontextmanager,
)

from ._utils import (
    cleanup_tasks,
    get_subprocess_command,
    read_exactly,
    receive_pickled_value,
)
from .abc import (
    ProcessAPI,
)
from .constants import (
    SIGINT_TIMEOUT_SECONDS,
    SIGTERM_TIMEOUT_SECONDS,
)
from .exceptions import (
    InvalidState,
)
from .process import (
    Process,
)
from .state import (
    State,
)
from .typing import (
    TReturn,
)

if TYPE_CHECKING:
    from typing import Tuple  # noqa: F401
    from .typing import SubprocessKwargs  # noqa: F401


logger = logging.getLogger("asyncio_run_in_process")


async def _monitor_sub_proc(
    proc: ProcessAPI[TReturn], sub_proc: asyncio.subprocess.Process, parent_w: int
) -> None:
    logger.debug("starting subprocess to run %s", proc)

    await proc.wait_pid()
    if proc.pid != sub_proc.pid:
        raise Exception("Process id mismatch.  This should not be possible")
    logger.debug("subprocess for %s started.  pid=%d", proc, sub_proc.pid)

    # we write the execution data immediately without waiting for the
    # `WAIT_EXEC_DATA` state to ensure that the child process doesn't have
    # to wait for that data due to the round trip times between processes.
    logger.debug("writing execution data for %s over fd=%d", proc, parent_w)
    # pass the child process the serialized `async_fn` and `args`

    with os.fdopen(parent_w, "wb") as to_child:
        to_child.write(proc.sub_proc_payload)
        to_child.flush()

    await proc.wait_for_state(State.WAIT_EXEC_DATA)
    logger.debug("child process %s waiting for exec data", proc)

    await proc.wait_for_state(State.STARTED)
    logger.debug("waiting for process %s finish", proc)

    await sub_proc.wait()

    proc.returncode = sub_proc.returncode
    logger.debug("process %s finished: returncode=%d", proc, proc.returncode)


async def _relay_signals(
    proc: ProcessAPI[Any],
    queue: 'asyncio.Queue[signal.Signals]',
) -> None:
    if proc.state.is_before(State.EXECUTING):
        # If the process has not reached the state where the child process
        # can properly handle the signal, give it a moment to reach the
        # `EXECUTING` stage.
        await proc.wait_for_state(State.EXECUTING)
    elif proc.state.is_on_or_after(State.STOPPING):
        await proc.wait_for_state(State.FINISHED)
        return

    while True:
        signum = await queue.get()

        logger.debug("relaying signal %s to child process %s", signum, proc)
        proc.send_signal(signum)


async def _monitor_state(
    proc: ProcessAPI[TReturn], parent_read_fd: int, loop: asyncio.AbstractEventLoop,
) -> None:
    with os.fdopen(parent_read_fd, "rb", closefd=True) as from_child:
        for expected_state in State:
            if proc.state is not expected_state:
                raise InvalidState(
                    f"Process in state {proc.state} but expected state {expected_state}"
                )

            child_state_as_byte = await loop.run_in_executor(None, read_exactly, from_child, 1)

            try:
                child_state = State(ord(child_state_as_byte))
            except TypeError:
                raise InvalidState(f"Child sent state: {child_state_as_byte.hex()}")

            if not proc.state.is_next(child_state):
                if proc.state is State.FINISHED:
                    # This case covers when the process is killed with a SIGKILL
                    # and jumps directly to the finished state.
                    return
                raise InvalidState(
                    f"Invalid state transition: {proc.state} -> {child_state}"
                )

            if child_state is State.FINISHED:
                # For the FINISHED state we delay updating the state until we also
                # have a return value.
                break
            elif child_state is State.INITIALIZED:
                # For the INITIALIZED state we expect an additional payload of the
                # process id.  The process ID is gotten via this mechanism to
                # prevent the need for ugly sleep based code in
                # `_monitor_sub_proc`.
                pid_bytes = await loop.run_in_executor(None, read_exactly, from_child, 4)
                proc.pid = int.from_bytes(pid_bytes, 'big')

            await proc.update_state(child_state)
            logger.debug(
                "Updated process %s state %s -> %s",
                proc,
                expected_state.name,
                child_state.name,
            )

        # This is mostly a sanity check but it ensures that the loop variable is
        # what we expect it to be before starting to collect the result the stream.
        if child_state is not State.FINISHED:
            raise InvalidState(f"Invalid final state: {proc.state}")

        result = await loop.run_in_executor(None, receive_pickled_value, from_child)

    await proc.wait_returncode()

    if proc.returncode == 0:
        proc.return_value = result
    else:
        proc.error = result

    await proc.update_state(child_state)
    logger.debug(
        "Updated process %s state %s -> %s",
        proc,
        expected_state.name,
        child_state.name,
    )


# SIGINT isn't included here because it's handled by catching the
# `KeyboardInterrupt` exception.
RELAY_SIGNALS = (signal.SIGTERM, signal.SIGHUP)


def open_in_process(
    async_fn: Callable[..., TReturn],
    *args: Any,
    loop: asyncio.AbstractEventLoop = None,
    subprocess_kwargs: 'SubprocessKwargs' = None,
) -> AsyncContextManager[ProcessAPI[TReturn]]:
    return cast(
        AsyncContextManager[ProcessAPI[TReturn]],
        _open_in_process(async_fn, *args, loop=loop, subprocess_kwargs=subprocess_kwargs),
    )


def _update_subprocess_kwargs(subprocess_kwargs: Optional['SubprocessKwargs'],
                              child_r: int,
                              child_w: int) -> 'SubprocessKwargs':
    if subprocess_kwargs is None:
        subprocess_kwargs = {}

    base_pass_fds = subprocess_kwargs.get('pass_fds', ())
    pass_fds: Tuple[int, ...]

    if base_pass_fds is None:
        pass_fds = (child_r, child_w)
    else:
        pass_fds = tuple(set(base_pass_fds).union((child_r, child_w)))

    updated_kwargs = subprocess_kwargs.copy()
    updated_kwargs['pass_fds'] = pass_fds

    return updated_kwargs


# mypy recognizes this decorator as being untyped.
@asynccontextmanager  # type: ignore
async def _open_in_process(
    async_fn: Callable[..., TReturn],
    *args: Any,
    loop: asyncio.AbstractEventLoop = None,
    subprocess_kwargs: 'SubprocessKwargs' = None,
) -> AsyncIterator[ProcessAPI[TReturn]]:
    proc: Process[TReturn] = Process(async_fn, args)

    parent_r, child_w = os.pipe()
    child_r, parent_w = os.pipe()
    parent_pid = os.getpid()

    command = get_subprocess_command(child_r, child_w, parent_pid)

    sub_proc = await asyncio.create_subprocess_exec(
        *command,
        **_update_subprocess_kwargs(subprocess_kwargs, child_r, child_w),
    )
    if loop is None:
        loop = asyncio.get_event_loop()

    signal_queue: asyncio.Queue[signal.Signals] = asyncio.Queue()

    for signum in RELAY_SIGNALS:
        loop.add_signal_handler(
            signum,
            signal_queue.put_nowait,
            signum,
        )

    # Monitoring
    monitor_sub_proc_task = asyncio.ensure_future(_monitor_sub_proc(proc, sub_proc, parent_w))
    relay_signals_task = asyncio.ensure_future(_relay_signals(proc, signal_queue))
    monitor_state_task = asyncio.ensure_future(_monitor_state(proc, parent_r, loop))

    async with cleanup_tasks(monitor_sub_proc_task, relay_signals_task, monitor_state_task):
        await proc.wait_pid()

        # Wait until the child process has reached the STARTED
        # state before yielding the context.  This ensures that any
        # calls to things like `terminate` or `kill` will be handled
        # properly in the child process.
        #
        # The timeout ensures that if something is fundamentally wrong
        # with the subprocess we don't hang indefinitely.
        await proc.wait_for_state(State.STARTED)

        try:
            try:
                yield proc
            except KeyboardInterrupt as err:
                # If a keyboard interrupt is encountered relay it to the
                # child process and then give it a moment to cleanup before
                # re-raising
                logger.debug("Relaying SIGINT to pid=%d", sub_proc.pid)
                try:
                    proc.send_signal(signal.SIGINT)
                    try:
                        await asyncio.wait_for(proc.wait(), timeout=SIGINT_TIMEOUT_SECONDS)
                    except asyncio.TimeoutError:
                        logger.debug(
                            "Timed out waiting for pid=%d to exit after relaying SIGINT",
                            sub_proc.pid,
                        )
                finally:
                    raise err
            except asyncio.CancelledError as err:
                # If this is due to an `asyncio` cancellation we send along a
                # SIGTERM to signal the need for more immediate cleanup, giving
                # the child process a moment to finish before re-raising.
                logger.debug(
                    "Got CancelledError while running subprocess pid=%d.  Sending SIGTERM.",
                    sub_proc.pid,
                )
                try:
                    proc.terminate()
                    try:
                        await asyncio.wait_for(proc.wait(), timeout=SIGTERM_TIMEOUT_SECONDS)
                    except asyncio.TimeoutError:
                        logger.debug(
                            "Timed out waiting for pid=%d to exit after SIGTERM",
                            sub_proc.pid,
                        )
                finally:
                    raise err
            else:
                # In the case that the yielded context block exits without an
                # error we wait for the process to finish naturally.  This can
                # hang indefinitely.
                await proc.wait()
        finally:
            if sub_proc.returncode is None:
                # If the process has not returned at this stage we need to hard
                # kill it to prevent it from hanging.
                logger.warning(
                    "Child process pid=%d failed to exit cleanly.  Sending SIGKILL",
                    sub_proc.pid,
                    # The `any` call is to include a stacktrace if this
                    # happened due to an exception but to omit it if this is
                    # somehow happening outside of an exception context.
                    exc_info=any(sys.exc_info()),
                )
                sub_proc.kill()


async def run_in_process(async_fn: Callable[..., TReturn],
                         *args: Any,
                         loop: asyncio.AbstractEventLoop = None,
                         subprocess_kwargs: 'SubprocessKwargs' = None) -> TReturn:
    proc_ctx = open_in_process(
        async_fn,
        *args,
        loop=loop,
        subprocess_kwargs=subprocess_kwargs,
    )
    async with proc_ctx as proc:
        await proc.wait()
    return proc.result



try:
    # mypy knows that `TypedDict` and `Literal` don't exist in <3.8
    from typing import TypedDict, Literal  # type: ignore
except ImportError:
    # TypedDict is only available in python 3.8+
    # Literal is only available in python 3.8+
    from typing_extensions import TypedDict, Literal  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)

extras_require = {
    'test': [
        "pytest==5.2.2",
        "pytest-asyncio==0.10.0",
        "pytest-xdist",
        "tox>=2.9.1,<3",
    ],
    'lint': [
        "flake8==3.4.1",
        "isort>=4.2.15,<5",
        "mypy==0.740",
        "pydocstyle>=3.0.0,<4",
        "typing-extensions>=3.7.4.1,<4;python_version<'3.8'",
    ],
    'doc': [
        "Sphinx>=1.6.5,<2",
        "sphinx_rtd_theme>=0.1.9",
    ],
    'dev': [
        "bumpversion>=0.5.3,<1",
        "pytest-watch>=4.1.0,<5",
        "wheel",
        "twine",
        "ipython",
    ],
}

extras_require['dev'] = (
    extras_require['dev'] +  # noqa: W504
    extras_require['test'] +  # noqa: W504
    extras_require['lint'] +  # noqa: W504
    extras_require['doc']
)


with open('./README.md') as readme:
    long_description = readme.read()


setup(
    name='asyncio-run-in-process',
    # *IMPORTANT*: Don't manually change the version here. Use `make bump`, as described in readme
    version='0.1.0-alpha.4',
    description="""asyncio-run-in-process: Asyncio friendly replacement for multiprocessing""",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='The Ethereum Foundation',
    author_email='snakecharmers@ethereum.org',
    url='https://github.com/ethereum/asyncio-run-in-process',
    include_package_data=True,
    install_requires=[
        "async-generator>=1.10,<2",
        "cloudpickle>=1.2.1,<2",
    ],
    python_requires='>=3.6, <4',
    extras_require=extras_require,
    py_modules=['asyncio_run_in_process'],
    license="MIT",
    zip_safe=False,
    keywords='ethereum',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)



import argparse
import asyncio
import logging
import os
import signal
import sys
from typing import (
    Any,
    BinaryIO,
    Callable,
    Coroutine,
    Sequence,
    cast,
)

from ._utils import (
    RemoteException,
    cleanup_tasks,
    pickle_value,
    receive_pickled_value,
)
from .state import (
    State,
)
from .typing import (
    TReturn,
)

logger = logging.getLogger("asyncio_run_in_process")


def update_state(to_parent: BinaryIO, state: State) -> None:
    to_parent.write(state.value.to_bytes(1, 'big'))
    to_parent.flush()


def update_state_initialized(to_parent: BinaryIO) -> None:
    payload = State.INITIALIZED.value.to_bytes(1, 'big') + os.getpid().to_bytes(4, 'big')
    to_parent.write(payload)
    to_parent.flush()


def update_state_finished(to_parent: BinaryIO, finished_payload: bytes) -> None:
    payload = State.FINISHED.value.to_bytes(1, 'big') + finished_payload
    to_parent.write(payload)
    to_parent.flush()


SHUTDOWN_SIGNALS = {signal.SIGTERM}


async def _handle_SIGTERM(task: 'asyncio.Task[TReturn]', fut: 'asyncio.Future[int]') -> None:
    signum = await fut
    raise SystemExit(signum)


async def _handle_coro(coro: Coroutine[Any, Any, TReturn], got_SIGINT: asyncio.Event) -> TReturn:
    """
    Understanding this function requires some detailed knowledge of how
    coroutines work.

    The goal here is to run a coroutine function and wait for the result.
    However, if a SIGINT signal is received then we want to inject a
    `KeyboardInterrupt` exception into the running coroutine.

    Some additional nuance:

    - The `SIGINT` signal can happen multiple times and each time we want to
      throw a `KeyboardInterrupt` into the running coroutine which may choose to
      ignore the exception and continue executing.
    - When the `KeyboardInterrupt` hits the coroutine it can return a value
      which is sent using a `StopIteration` exception.  We treat this as the
      return value of the coroutine.
    """
    # The `coro` is first wrapped in `asyncio.shield` to protect us in the case
    # that a SIGINT happens.  In this case, the coro has a chance to return a
    # value during exception handling, in which case we are left with
    # `coro_task` which has not been awaited and thus will cause asyncio to
    # issue a warning.  However, since the coroutine has already exited, if we
    # await the `coro_task` then we will encounter a `RuntimeError`.  By
    # wrapping the coroutine in `asyncio.shield` we side-step this by
    # preventing the cancellation to actually penetrate the coroutine, allowing
    # us to await the `coro_task` without causing the `RuntimeError`.
    coro_task = asyncio.ensure_future(asyncio.shield(coro))
    async with cleanup_tasks(coro_task):
        while True:
            # Run the coroutine until it either returns, or a SIGINT is received.
            # This is done in a loop because the function *could* choose to ignore
            # the `KeyboardInterrupt` and continue processing, in which case we
            # reset the signal and resume waiting.
            done, pending = await asyncio.wait(
                (coro_task, got_SIGINT.wait()),
                return_when=asyncio.FIRST_COMPLETED,
            )

            if coro_task.done():
                async with cleanup_tasks(*done, *pending):
                    return await coro_task
            elif got_SIGINT.is_set():
                got_SIGINT.clear()

                # In the event that a SIGINT was recieve we need to inject a
                # KeyboardInterrupt exception into the running coroutine.
                try:
                    coro.throw(KeyboardInterrupt)
                except StopIteration as err:
                    # StopIteration is how coroutines signal their return values.
                    # If the exception was raised, we treat the argument as the
                    # return value of the function.
                    async with cleanup_tasks(*done, *pending):
                        return cast(TReturn, err.value)
                except BaseException as err:
                    raise
            else:
                raise Exception("Code path should not be reachable")


async def _do_async_fn(
    async_fn: Callable[..., Coroutine[Any, Any, TReturn]],
    args: Sequence[Any],
    to_parent: BinaryIO,
    loop: asyncio.AbstractEventLoop,
) -> TReturn:
    # state: STARTED
    update_state(to_parent, State.STARTED)

    # A Future that will be set if any of the SHUTDOWN_SIGNALS signals are
    # received causing _do_async_fn to raise a SystemExit
    system_exit_signum: 'asyncio.Future[int]' = asyncio.Future()

    # setup signal handlers.
    for signum in SHUTDOWN_SIGNALS:
        loop.add_signal_handler(
            signum.value,
            system_exit_signum.set_result,
            signum,
        )

    # state: EXECUTING
    update_state(to_parent, State.EXECUTING)

    # Install a signal handler to set an asyncio.Event upon receiving a SIGINT
    got_SIGINT = asyncio.Event()
    loop.add_signal_handler(
        signal.SIGINT,
        got_SIGINT.set,
    )

    # First we need to generate a coroutine.  We need this so we can throw
    # exceptions into the running coroutine to allow it to handle keyboard
    # interrupts.
    async_fn_coro: Coroutine[Any, Any, TReturn] = async_fn(*args)

    # The coroutine is then given to `_handle_coro` which waits for either the
    # coroutine to finish, returning the result, or for a SIGINT signal at
    # which point injects a `KeyboardInterrupt` into the running coroutine.
    async_fn_task: 'asyncio.Future[TReturn]' = asyncio.ensure_future(
        _handle_coro(async_fn_coro, got_SIGINT),
    )

    # Now we wait for either a result from the coroutine or a SIGTERM which
    # triggers immediate cancellation of the running coroutine.
    done, pending = await asyncio.wait(
        (async_fn_task, system_exit_signum),
        return_when=asyncio.FIRST_COMPLETED,
    )

    # We prioritize the `SystemExit` case.
    async with cleanup_tasks(*done, *pending):
        if system_exit_signum.done():
            raise SystemExit(await system_exit_signum)
        elif async_fn_task.done():
            return await async_fn_task
        else:
            raise Exception("unreachable")


def _run_process(parent_pid: int, fd_read: int, fd_write: int) -> None:
    """
    Run the child process.

    This communicates the status of the child process back to the parent
    process over the given file descriptor, runs the coroutine, handles error
    cases, and transmits the result back to the parent process.
    """
    # state: INITIALIZING (default initial state)
    with os.fdopen(fd_write, "wb") as to_parent:
        # state: INITIALIZED
        update_state_initialized(to_parent)

        with os.fdopen(fd_read, "rb", closefd=True) as from_parent:
            # state: WAIT_EXEC_DATA
            update_state(to_parent, State.WAIT_EXEC_DATA)
            async_fn, args = receive_pickled_value(from_parent)

        # state: BOOTING
        update_state(to_parent, State.BOOTING)

        loop = asyncio.get_event_loop()

        try:
            try:
                result = loop.run_until_complete(
                    _do_async_fn(async_fn, args, to_parent, loop),
                )
            except BaseException:
                _, exc_value, exc_tb = sys.exc_info()
                # `mypy` thingks that `exc_value` and `exc_tb` are `Optional[..]` types
                remote_exc = RemoteException(exc_value, exc_tb)  # type: ignore
                finished_payload = pickle_value(remote_exc)
                raise
            finally:
                # state: STOPPING
                update_state(to_parent, State.STOPPING)
        except KeyboardInterrupt:
            code = 2
        except SystemExit as err:
            code = err.args[0]
        except BaseException:
            code = 1
        else:
            finished_payload = pickle_value(result)
            code = 0
        finally:
            # state: FINISHED
            update_state_finished(to_parent, finished_payload)
            sys.exit(code)


#
# CLI invocation for subprocesses
#
parser = argparse.ArgumentParser(description="asyncio-run-in-process")
parser.add_argument(
    "--parent-pid", type=int, required=True, help="The PID of the parent process"
)
parser.add_argument(
    "--fd-read",
    type=int,
    required=True,
    help=(
        "The file descriptor that the child process can use to read data that "
        "has been written by the parent process"
    ),
)
parser.add_argument(
    "--fd-write",
    type=int,
    required=True,
    help=(
        "The file descriptor that the child process can use for writing data "
        "meant to be read by the parent process"
    ),
)


if __name__ == "__main__":
    args = parser.parse_args()
    _run_process(
        parent_pid=args.parent_pid, fd_read=args.fd_read, fd_write=args.fd_write
    )



from .exceptions import (  # noqa: F401
    BaseRunInProcessException,
    InvalidState,
    ProcessKilled,
)
from .run_in_process import (  # noqa: F401
    open_in_process,
    run_in_process,
)
from .state import (  # noqa: F401
    State,
)



# The number of seconds that are given to a child process to exit after the
# parent process gets a KeyboardInterrupt/SIGINT-signal and sends a `SIGINT` to
# the child process.
SIGINT_TIMEOUT_SECONDS = 2


# The number of seconds that are givent to a child process to exit after the
# parent process gets an `asyncio.CancelledError` which results in sending a
# `SIGTERM` to the child process.
SIGTERM_TIMEOUT_SECONDS = 2
